{$ObjExportAll On}
unit DirectoryTreeView;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls;

type
  TDirectoryTreeView = class(TCustomTreeView)
  private
    { Private declarations }
    FImages: TImageList;
    FOnChange: TNotifyEvent;
    FDirectory: String;
    FSetDirectory: Boolean;
    FRootRead: Boolean;
    procedure ExpandLevel( N: TTreeNode; P: String);
    procedure Deletion( Sender: TObject; Node: TTreeNode);
    procedure SetDirectory( Value: String);
    function GetDirectory: String;
    function GetVersion: String;
    procedure SetVersion( Value: String);
    function SplitDirectory( P: String): TStringList;
  protected
    { Protected declarations }
    procedure Expand( Node: TTreeNode); override;
    procedure CreateWnd; override;
    function CanExpand(Node: TTreeNode): Boolean; override;
    procedure Change(Node: TTreeNode); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Refresh;
    procedure FindDirectory( P: String);
  published
    { Published declarations }
    property Directory: String read GetDirectory write SetDirectory;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property Version: String read GetVersion write SetVersion;

    property ShowButtons;
    property BorderStyle;
    property DragCursor;
    property ShowLines;
    property ShowRoot;
    property RightClickSelect;
    property DragMode;
    property HideSelection;
    property Indent;
    property Align;
    property Enabled;
    property Font;
    property Color;
    property ParentColor default False;
    property ParentCtl3D;
    property Ctl3D;
    property SortType;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnEnter;
    property OnExit;
    property OnClick;
    property OnDragDrop;
    property OnDragOver;
    property OnStartDrag;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnDblClick;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property PopupMenu;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
   end;

procedure Register;

implementation

{$R 'DirectoryTreeView.res'}

procedure TDirectoryTreeView.SetVersion( Value: String);
begin
  { Dummy procedure }
end;

function TDirectoryTreeView.GetVersion: String;
begin
  Result := '1.04';
end;

procedure TDirectoryTreeView.Change(Node: TTreeNode);
begin
  if Assigned( FOnChange) then FOnChange( Self);
end;

function TDirectoryTreeView.CanExpand(Node: TTreeNode): Boolean;
begin
  if csDesigning in ComponentState then
     Result := False else
     Result := inherited CanExpand( Node);
end;

procedure TDirectoryTreeView.SetDirectory( Value: String);
begin
  FindDirectory( Value);
end;

function TDirectoryTreeView.GetDirectory: String;
begin
  Result := '';
  if Selected <> nil then
     Result := PChar( Selected.Data)+1;
end;

function TDirectoryTreeView.SplitDirectory( P: String): TStringList;
var
  D: TStringList;
  C: Integer;
begin
  D := TStringList.Create;

  for C := 1 to Length( P) do
  if (P[C] = '\') or (P[C] = '/') then
  begin
    if P[C] = '/' then
       P[C] := '\';
       
    D.Add( Copy( P, 1, C));
  end;

  Result := D;
end;

procedure TDirectoryTreeView.FindDirectory( P: String);
var
  D: TStringList;
  C: Integer;
  I: Integer;
  R: TTreeNode;
begin
  if P = '' then
     Exit;

  if (Items.Count <= 0) or (not FRootRead) then
  begin
    FDirectory := P;
    FSetDirectory := True;
    Exit;
  end;

  if (P[ Length( P)-1] <> '\') and (P[ Length( P)-1] <> '/') then
     P := P + '\';

  { Split the directory up in subdir strings }
  D := SplitDirectory(P);
  try
    { Start searching for the subdirs }
    for I := 0 to D.Count-1 do
    begin
      R := nil;

      { Find node(Subdir) }
      for C := 0 to Items.Count-1 do
      begin
        { Expand node(Subdir) }
        if UpperCase( PChar( Items.Item[C].Data)+1) = UpperCase( D[I]) then
        begin
          ExpandLevel( Items.Item[C], '');
          R := Items.Item[C];
          Break;
        end;
      end;

      { node(Subdir) not found }
      if R = nil then
         Exit;

      { Make the node(Subdir) active }
      Selected := R;
    end;
  finally
    { Free the subdir list }
    D.Free;
  end;
end;

procedure TDirectoryTreeView.Deletion( Sender: TObject; Node: TTreeNode);
begin
  { Free the directory data }
  if Node <> nil then
  if Node.Data <> nil then
     FreeMem( Node.Data);
end;

procedure TDirectoryTreeView.Refresh;
var
  D: DWORD;
  C: Integer;
  V: Array [0..512] of Char;
  L: String;
  R: Array [0..3] of Char;
  H: DWORD;
  Dir: String;
begin
  Dir := Directory;

  { Clear all the directories }
  Items.Clear;

  { Get all logical drives }
  D := GetLogicalDrives;

  { Skip A: & B: }
  D := D shr 2;

  { Loop until all the drives have been added }
  for C := 3 to 32 do
  begin
    { Is a drive found ? }
    if D and 1 = 1 then
    begin
      { Yes, add it to the treeview }
      StrPCopy( R, Char(Byte('A')+C-1)+':\');

      { What kind of drive is it ? }
      V[0] := Char(0);
      GetVolumeInformation( R, V, 512, nil, H, H, nil, 0);
      L := V;
      if Length( L) <> 0 then
         L := L + ' ';

      { Get all the subdirs in the root of each drive }
      ExpandLevel( Items.Add( nil, L+'('+Char(Byte('A')+C-1)+':)'), Char(Byte('A')+C-1)+':\');
    end;

    { Check the next drive }
    D := D shr 1;
  end;

  Directory := Dir;
  FRootRead := True;
end;

procedure TDirectoryTreeView.Expand( Node: TTreeNode);
var
  C: Integer;
begin
  inherited;
  { Get all the subdirectories of the selected node }
  for C := 0 to Node.Count - 1 do
      ExpandLevel( Node.Item[C], '');
end;

constructor TDirectoryTreeView.Create(AOwner: TComponent);
begin
  inherited Create( AOwner);

  { Add an OnDelete event to make sure everything gets free'd upon destroy }
  OnDeletion := Deletion;

  { Make the treeview readonly, we can't have users rename the directories.
    Atleast not yet. }
  ReadOnly := True;

  HideSelection := False;
  FRootRead := False;

  FImages := TImageList.Create( Self);

  FImages.ResInstLoad( HInstance, rtBitmap, 'FLOPPY', clBlue);
  FImages.ResInstLoad( HInstance, rtBitmap, 'HARDDRIVE', clBlue);
  FImages.ResInstLoad( HInstance, rtBitmap, 'NETWORK', clBlue);
  FImages.ResInstLoad( HInstance, rtBitmap, 'CDROM', clBlue);
  FImages.ResInstLoad( HInstance, rtBitmap, 'FOLDERCLOSED', clBlue);
  FImages.ResInstLoad( HInstance, rtBitmap, 'CURRENTFOLDER', clBlue);

  Images := FImages;
  StateImages := FImages;
end;

destructor TDirectoryTreeView.Destroy;
begin
  FImages.Free;
  inherited Destroy;
end;

procedure TDirectoryTreeView.CreateWnd;
begin
  inherited CreateWnd;
  Refresh;

  if FSetDirectory then
     Directory := FDirectory;
end;

procedure TDirectoryTreeView.ExpandLevel( N: TTreeNode; P: String);
var
  S: PChar;
  F: TSearchRec;
  R: Integer;
  C: Integer;
  D: TStringList;
  CN: TTreeNode;
begin
  if N.Data = nil then
  begin
    GetMem( S, Length( P)+2);
    StrPCopy( S+1, P);
    S[0] := #0;
    N.Data := S;
  end;

  P := PChar( N.Data)+1;
  if PChar( N.Data)[0] = #0 then
  begin
    D := TStringList.Create;
    try
      R := FindFirst( P+'*.*', faDirectory, F);
      while R = 0 do
      begin
        if F.Attr and faDirectory <> 0 then
        if F.Name <> '.' then
        if F.Name <> '..' then
        begin
          D.Add( F.Name);
        end;
        R := FindNext( F);
      end;
      FindClose( F);

      D.Sorted := True;
      for C := 0 to D.Count-1 do
      begin
        if N.HasChildren then
           CN := Items.AddChild( N, D[C]) else
           CN := Items.AddChildFirst( N, D[C]);

        with CN do
        begin
          GetMem( S, Length( P+D[C]+'\')+2);
          StrPCopy( S+1, P+D[C]+'\');
          S[0] := #0;
          Data := S;

          ImageIndex := 4;
          SelectedIndex := 5;
        end;
      end;
    finally
      D.Free;
    end;
    PChar( N.Data)[0] := #1;

    if N.Level = 0 then
    begin
      case GetDriveType( PChar( N.Data)+1) of
        DRIVE_UNKNOWN, DRIVE_NO_ROOT_DIR, DRIVE_RAMDISK: N.ImageIndex := 1;
        DRIVE_FIXED: N.ImageIndex := 1;
        DRIVE_REMOVABLE: N.ImageIndex := 0;
        DRIVE_REMOTE: N.ImageIndex := 2;
        DRIVE_CDROM: N.ImageIndex := 3;
      end;
      N.SelectedIndex := N.ImageIndex;
    end;
  end;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TDirectoryTreeView]);
end;

end.
