#ifndef _memmap_h_
#define _memmap_h_

#include "snes9x.h"

#ifdef FAST_LSB_WORD_ACCESS
#define READ_WORD(s) (*(uint16 *) (s))
#define WRITE_WORD(s, d) (*(uint16 *) (s) = (d)
#else
#define READ_WORD(s) (*(uint8 *) (s) | (*((uint8 *) (s) + 1) << 8))
#define WRITE_WORD(s, d) *(uint8 *) (s) = (d), \
                         *((uint8 *) (s) + 1) = (d) >> 8
#endif

class CMemory {
public:
    bool8 LoadROM (const char *);
    bool8 LoadSRAM (const char *);
    bool8 SaveSRAM (const char *);
    bool8 Init ();
    void  Deinit ();

    void WriteProtectROM ();
    void FixROMSpeed ();
    void MapRAM ();
    char *Safe (const char *);
    
    void LoROMMap ();
    void HiROMMap ();
    void SuperFXROMMap ();
    void TalesROMMap ();
    void AlphaROMMap ();
    void SA1ROMMap ();
    bool8 AllASCII (uint8 *b, int size);
    int  ScoreHiROM (bool8 skip_header);
    int  ScoreLoROM (bool8 skip_header);
    void ApplyROMFixes ();
    const char *TVStandard ();
    const char *Speed ();
    const char *StaticRAMSize ();
    const char *MapType ();
    const char *MapMode ();
    const char *KartContents ();
    const char *Size ();
    const char *Headers ();
    const char *ROMID ();

    enum {
	MAP_PPU, MAP_CPU, MAP_DSP, MAP_LOROM_SRAM, MAP_HIROM_SRAM,
	MAP_NONE, MAP_DEBUG, MAP_LAST
    };
//    enum { MAX_ROM_SIZE = 0x400000 };
    enum { MAX_ROM_SIZE = 0x600000 };
    
    uint8 *RAM;
    uint8 *ROM;
    uint8 *VRAM;
    uint8 *SRAM;
    uint8 *FillRAM;
    bool8 HiROM;
    bool8 LoROM;
    uint16 SRAMMask;
    uint8 SRAMSize;
    uint8 *Map [256 * 8];
    uint8 *WriteMap [256 * 8];
    uint8 MemorySpeed [256 * 8];
    uint8 BlockIsRAM [256 * 8];
    uint8 BlockIsROM [256 * 8];
    char  ROMName [ROM_NAME_LEN];
    char  ROMId [5];
    uint8 ROMSpeed;
    uint8 ROMType;
    uint8 ROMSize;
    int32 ROMFramesPerSecond;
    int32 HeaderCount;
    uint32 CalculatedSize;
    uint32 CalculatedChecksum;
    uint32 ROMChecksum;
    uint32 ROMComplementChecksum;
    char ROMFilename [_MAX_PATH];
};

START_EXTERN_C
extern CMemory Memory;
extern uint8 *SRAM;
extern uint8 *ROM;
extern uint8 *RegRAM;
END_EXTERN_C

#ifdef NO_INLINE_SET_GET
uint8 S9xGetByte (uint32 Address);
uint16 S9xGetWord (uint32 Address);
void S9xSetByte (uint8 Byte, uint32 Address);
void S9xSetWord (uint16 Byte, uint32 Address);
void S9xSetPCBase (uint32 Address);
uint8 *S9xGetMemPointer (uint32 Address);
uint8 *GetBasePointer (uint32 Address);
#else
#define INLINE inline
#include "getset.h"
#endif // NO_INLINE_SET_GET

#endif // _memmap_h_
