#ifndef _NETPLAY_H_
#define _NETPLAY_H_

/* Server to client header format:
 *
 * magic	4
 * version	1
 * opcode	1
 * len		4
 * sequence_no	4
 * joypads	4 * 5
 * additional data...
 */

 /* Client to server packet format:
  * magic	4
  * version	1
  * opcode	1
  * len		4
  * data	?
  */
  
typedef struct {
    char   magic [4];
    uint8  version;
    uint8  opcode;
    uint32 len;
    uint32 sequence_no;
} SCNetPlay;

#define NETPLAY_SERVER_VERSION_OFFSET 4
#define NETPLAY_SERVER_OPCODE_OFFSET 5
#define NETPLAY_SERVER_LEN_OFFSET 6
#define NETPLAY_SERVER_SEQUENCE_NO_OFFSET 10
#define NETPLAY_SERVER_JOYPAD_OFFSET 14
#define NETPLAY_SERVER_TO_CLIENT_HEADER_SIZE (4 + 1 + 1 + 4 + 4)

typedef struct {
    char   magic [4];
    uint8  version;
    uint8  opcode;
    uint32 len;
    uint32 sequence_no;
} CSNetPlay;

#define NETPLAY_MAGIC "S9X"
#define NETPLAY_VERSION 1
#define NETPLAY_CLIENT_TO_SERVER_HEADER_SIZE (4 + 1 + 1 + 4)
#define NETPLAY_CLIENT_VERSION_OFFSET 4
#define NETPLAY_CLIENT_OPCODE_OFFSET 5
#define NETPLAY_CLIENT_LEN_OFFSET 6
#define NETPLAY_CLIENT_SEQUENCE_NO_OFFSET 10

#define NETPLAY_DEFAULT_PORT 6096

/* Server to client */
enum {
    NETPLAY_SERVER_HELLO,
    NETPLAY_SERVER_JOYPAD_UPDATE,
    NETPLAY_SERVER_RESET,
    NETPLAY_SERVER_PAUSE,
    NETPLAY_SERVER_ROM_NAME
};

/* Client to server */
enum {
    NETPLAY_CLIENT_HELLO,
    NETPLAY_CLIENT_JOYPAD_UPDATE,
    NETPLAY_CLIENT_GOODBYE
};

//
// NETPLAY_CLIENT_HELLO message format:
// header
// frame_time (4)
// ROMName (variable)

#define WRITE_LONG(p, v) { \
*((p) + 0) = (uint8) (v >> 24); \
*((p) + 1) = (uint8) (v >> 16); \
*((p) + 2) = (uint8) (v >> 8); \
*((p) + 3) = (uint8) (v >> 0); \
}

#define READ_LONG(p, v) \
v = ((uint8) *((p) + 0) << 24) | \
    ((uint8) *((p) + 1) << 16) | \
    ((uint8) *((p) + 2) <<  8) | \
    ((uint8) *((p) + 3) <<  0)

bool8 S9xNetPlayConnectToServer (const char *server_name, int port,
				 const char *rom_name, int &player_number_return);
bool8 S9xNetPlayCheckForHeartBeat ();
uint32 S9xNetPlayGetJoypad (int which1);
bool8 S9xNetPlaySendJoypadUpdate (uint32 joypad);
void S9xNetPlayDisconnect ();

bool8 S9xNetPlaySend (void *fd, const char *data, int len);
char *S9xNetPlayReceivePacket (void *fd, int &len);
#endif
