#include "snes9x.h"
#include "..\directx.h"
#include "memmap.h"
#include "debug.h"
#include "cpuexec.h"
#include "ppu.h"
#include "snapshot.h"
#include "apu.h"
#include "display.h"
#include "gfx.h"
#include "soundux.h"

struct SJoyState{
    bool Attached;
    JOYCAPS Caps;
    int Threshold;
    bool Left;
    bool Right;
    bool Up;
    bool Down;
    bool PovLeft;
    bool PovRight;
    bool PovUp;
    bool PovDown;
    bool Button[32];
} Joystick[16];

FILE *trace_fs = NULL;

int __fastcall Normalize( int cur, int min, int max)
{
int Result;
    Result = 0;

    if( (max - min) == 0)
        return Result;

    Result = cur - min;
    Result = (Result*200) / (max - min);
    Result = Result - 100;

    return Result;
}

void S9xTextMode( void)
{
}

void S9xGraphicsMode ()
{
}

char rom_name [ROM_NAME_LEN];

void S9xExit( void)
{
	SendMessage( GUI.hWnd, WM_COMMAND, ID_FILE_EXIT, 0);
}

extern "C" const char *S9xGetSRAMFilename ()
{
    static char filename [_MAX_PATH];
    char drive [_MAX_DRIVE];
    char dir [_MAX_DIR];
    char fname [_MAX_FNAME];
    char ext [_MAX_EXT];

    _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
    _makepath( filename, drive, dir, fname, ".srm");
    return (filename);
}

bool8 S9xOpenSnapshotFile( const char *fname, bool8 read_only, STREAM *file)
{
    char filename [_MAX_PATH];
    char drive [_MAX_DRIVE];
    char dir [_MAX_DIR];
    char fn [_MAX_FNAME];
    char ext [_MAX_EXT];

    _splitpath( fname, drive, dir, fn, ext);
    _makepath( filename, drive, dir, fn, ext[0] == '\0' ? ".000" : ext);

    if (read_only)
    {
	if ((*file = OPEN_STREAM (filename, "rb")))
	    return (TRUE);
    }
    else
    {
	if ((*file = OPEN_STREAM (filename, "wb")))
	    return (TRUE);
    }
    return (FALSE);
}

void S9xCloseSnapshotFile( STREAM file)
{
    CLOSE_STREAM (file);
}

void S9xMessage (int, int, const char *str)
{
#ifdef DEBUGGER
    static FILE *out = NULL;

    if (out == NULL)
        out = fopen ("out.txt", "w");

    fprintf (out, "%s\n", str);
#endif

#if 1
	fprintf (trace_fs, "%s\n", str);
	fflush (trace_fs);
#endif

	S9xSetInfoString (str);
	DirectX.OnScreenMsg( (char *)str, 5000);
}

long _rate = 44000;
unsigned long _interval = 5;
bool _sixteen_bit = true;
bool _stereo = true;
long _buffernos = 20;
long _buffersize = 4400;
long _samplecount = 440;
long _bytecount = 880;
long _buffersizemask = 0;
uint8 *SoundBuffer = NULL;
bool StartPlaying = false;
extern volatile unsigned long LastSound;
DWORD _lastblock = 1;

bool SetupSound( long rate, bool sixteen_bit, bool stereo)
{
    extern FILE *trace_fs;

    if (!trace_fs)
        trace_fs = fopen ("trace.log", "w+");

    so.mute_sound = TRUE;
    if( SoundBuffer != NULL)
        delete SoundBuffer;

    SoundBuffer = NULL;

    _rate = rate;
    _interval = 5;
    _sixteen_bit = sixteen_bit;
    _stereo = stereo;
    _buffernos = 2;
    int s = (rate * _interval * 2 * (_stereo ? 2 : 1) * (_sixteen_bit ? 2 : 1)) / 1000;

    _buffersize = 256;
    while (_buffersize < s)
        _buffersize *= 2;

    _samplecount = _sixteen_bit ? _buffersize / 2 : _buffersize;
    _bytecount = _buffersize;
    _buffersizemask = (_buffersize * _buffernos) - 1;
    _lastblock = 1;

    StartPlaying = false;

    so.playback_rate = _rate;
    so.stereo = _stereo;
    so.sixteen_bit = _sixteen_bit;
    so.buffer_size = _buffersize;
    so.encoded = FALSE;

    fprintf (trace_fs, "_buffersize = %d, so.buffer_size = %d,_rate = %d, _stereo = %d, _sixteen_bit = %d\n",
         _buffersize, so.buffer_size, _rate, _stereo, _sixteen_bit);
    fflush (trace_fs);

    if( DirectX.SetSoundMode())
        SoundBuffer = new uint8 [_buffersize*_buffernos];

    so.mute_sound = FALSE;
    LastSound = timeGetTime();
    return (SoundBuffer != NULL);
}

#define FIXED_POINT 0x10000
#define FIXED_POINT_SHIFT 16
#define FIXED_POINT_REMAINDER 0xffff

static bool8 block_signal = FALSE;
static volatile bool8 pending_signal = FALSE;

void ProcessSound ();

extern "C" void S9xGenerateSound(void)
{
    if (so.samples_mixed_so_far >= (uint32) _samplecount)
	return;

    block_signal = TRUE;
    so.err_counter += so.err_rate;
    if (so.err_counter >= FIXED_POINT)
    {
        int sample_count = so.err_counter >> FIXED_POINT_SHIFT;
	int byte_offset;
	int byte_count;

        so.err_counter &= FIXED_POINT_REMAINDER;
	if (so.stereo)
	    sample_count <<= 1;
        if (so.samples_mixed_so_far + sample_count > (uint32) _samplecount)
            sample_count = _samplecount - so.samples_mixed_so_far;

	byte_count = sample_count;

	if (so.sixteen_bit)
	{
	    byte_offset = (so.samples_mixed_so_far << 1) + so.play_position;
	    byte_count <<= 1;
	}
	else
	    byte_offset = so.samples_mixed_so_far + so.play_position;
	    
        S9xMixSamplesO (SoundBuffer, sample_count, 
			byte_offset & _buffersizemask);
        so.samples_mixed_so_far += sample_count;
    }
    block_signal = FALSE;
    if (pending_signal)
    {
	ProcessSound ();
	pending_signal = FALSE;
    }
}

// Interval ms has passed
void ProcessSound( void)
{
    bool8 block = false;

    if( DirectX.lpDSB == NULL)
    {
        fprintf (trace_fs, "ProcessSound DirectX.lpDSB == NULL\n");
        fflush (trace_fs);
        return;
    }

    if( SoundBuffer == NULL)
    {
        fprintf (trace_fs, "ProcessSound SoundBuffer == NULL\n");
        fflush (trace_fs);
        return;
    }

    if (block_signal)
    {
        pending_signal = TRUE;
        return;
    }
    if (block)
        return;

    block = true;
    if( StartPlaying)
    {
        fprintf (trace_fs, "ProcessSound StartPlaying\n");
        fflush (trace_fs);
        if (DirectX.lpDSB -> Play( 0, 0, DSBPLAY_LOOPING) != DS_OK)
            fprintf (trace_fs, "DirectSound Play failed\n");

        _lastblock = 1;
        StartPlaying = false;
    }
    DWORD play_pos = 0, write_pos = 0;
    HRESULT hResult;

    DirectX.lpDSB -> GetCurrentPosition (&play_pos, &write_pos);

//fprintf (trace_fs, "%d,", play_pos);
//fflush (trace_fs);

    if( play_pos / _buffersize != _lastblock)
    {	
	BYTE  *B1, *B2;
	DWORD S1, S2;

        _lastblock = play_pos / _buffersize;
        write_pos = (_lastblock + 1) % _buffernos * _buffersize;

        hResult = DirectX.lpDSB->Lock (write_pos, _buffersize, (void **)&B1, 
                                       &S1, (void **)&B2, &S2, 0);
        if (hResult == DSERR_BUFFERLOST)
        {
            fprintf (trace_fs, "ProcessSound BUFFERLOST\n");
            fflush (trace_fs);

            DirectX.lpDSB->Restore ();
            hResult = DirectX.lpDSB->Lock (write_pos, _buffersize, 
                                           (void **)&B1, &S1, (void **)&B2, 
                                           &S2, 0);
        }

        if( hResult != DS_OK)
        {
            fprintf (trace_fs, "ProcessSound Lock failed\n");
            fflush (trace_fs);
            block = false;
            return;
        }

        if( GUI.PauseEmulation || GUI.StopEmulation)
        {
            if (_sixteen_bit)
            {
                if (B1)
                    ZeroMemory (B1, S1);
                if (B2)
                    ZeroMemory (B2, S2);
            }
            else
            {
                if (B1)
                    memset (B1, 128, S1);
                if (B2)
                    memset (B2, 128, S2);
            }
        }
        else
        {
            unsigned int sample_count = so.buffer_size;
            int byte_offset;

            if (so.sixteen_bit)
                sample_count >>= 1;

            if (so.samples_mixed_so_far < sample_count)
            {
                byte_offset = so.play_position + 
                    (so.sixteen_bit ? (so.samples_mixed_so_far << 1)
                     : so.samples_mixed_so_far);

                if (Settings.SoundSync == 2)
                {
                    /*memset (SoundBuffer + (byte_offset & SOUND_BUFFER_SIZE_MASK), 0,
                      sample_count - so.samples_mixed_so_far);*/
                }
                else
                    S9xMixSamplesO (SoundBuffer, sample_count - so.samples_mixed_so_far,
                                    byte_offset & _buffersizemask);
                so.samples_mixed_so_far = 0;
            }
            else
                so.samples_mixed_so_far -= sample_count;
#if 1
            if (B1)
                memmove (B1, SoundBuffer + so.play_position, S1);
            if (B2)
                memmove (B2, SoundBuffer + so.play_position + S1, S2);
#else
            if (B1)
                S9xMixSamples (B1, _sixteen_bit ? (S1 >> 1) : S1);
            if (B2)
                S9xMixSamples (B2, _sixteen_bit ? (S2 >> 1) : S2);
#endif
            so.play_position = (so.play_position + so.buffer_size) & 
                               _buffersizemask;
        }
        hResult = DirectX.lpDSB -> Unlock( B1, S1, B2, S2);
        if( hResult != DS_OK)
        {
            fprintf (trace_fs, "ProcessSound Unlock failed\n");
            fflush (trace_fs);
            block = false;
            return;
        }
    }
    DWORD Status;
    hResult = DirectX.lpDSB -> GetStatus( &Status);
    if (StartPlaying = !Status)
    {
        fprintf (trace_fs, "ProcessSound need to StartPlaying\n");
        fflush (trace_fs);
    }
    block = false;
}
 

bool8 S9xOpenSoundDevice (int mode, bool8 pStereo, int BufferSize)
{	return (TRUE);	}

bool8 S9xReadSuperScopePosition (int &, int &, uint32 &)
{
	return (FALSE);
}

extern volatile long CurrentFrame;
extern volatile long TargetFrame;
void S9xSyncSpeed( void)
{
	CurrentFrame ++;
	if( Settings.SkipFrames == AUTO_FRAMERATE)
	{
		if( TargetFrame == CurrentFrame)
		{	
			IPPU.SkippedFrames =	0;
			IPPU.RenderThisFrame = TRUE;
		}
		else
		{	
			if( TargetFrame < CurrentFrame)
			{	
				IPPU.SkippedFrames =	0;
				IPPU.RenderThisFrame = TRUE;
				WaitForSingleObject( GUI.hDrawFrame, 100);
			}
			else
			{	
				IPPU.SkippedFrames ++;
				if( IPPU.SkippedFrames > 5)
				{
					IPPU.SkippedFrames =	0;
					IPPU.RenderThisFrame = TRUE;
				}
				else
				{	IPPU.RenderThisFrame = FALSE;	}
			}
		}
	}
	else
	{
		if( TargetFrame <= CurrentFrame)
		{
			IPPU.RenderThisFrame = TRUE;

			// For all you lamers out there thinking conspiritie theory things,
			// you'll notice that this version does not count the number
			// of frames skipped, but rather the no. of the frame it has to draw
			// So if Settings.SkipFrames == 0 then it should draw the next screen
			// thus it should draw SkipFrames + 1 !
			TargetFrame = CurrentFrame + Settings.SkipFrames + 1;
		}
		else
		{	IPPU.RenderThisFrame = FALSE;	}
	}
}

const char *basename (const char *f)
{
    const char *p;
    if ((p = strrchr (f, '/')) != NULL || (p = strrchr (f, '\\')) != NULL)
	return (p + 1);

#ifdef __DJGPP
    if (p = strrchr (f, SLASH_CHAR))
	return (p + 1);
#endif

    return (f);
}

bool8 S9xReadMousePosition (int which, int &x, int &y, DWORD &buttons)
{
	return (0);
}

bool GetState( WORD KeyIdent)
{
	if( KeyIdent&0x8000)
	{
		switch( KeyIdent&0xFF)
		{
		case 0: return !Joystick[ (KeyIdent>>8)^0x80].Left;
		case 1: return !Joystick[ (KeyIdent>>8)^0x80].Right;
		case 2: return !Joystick[ (KeyIdent>>8)^0x80].Up;
		case 3: return !Joystick[ (KeyIdent>>8)^0x80].Down;
		case 4: return !Joystick[ (KeyIdent>>8)^0x80].PovLeft;
		case 5: return !Joystick[ (KeyIdent>>8)^0x80].PovRight;
		case 6: return !Joystick[ (KeyIdent>>8)^0x80].PovUp;
		case 7: return !Joystick[ (KeyIdent>>8)^0x80].PovDown;
		default:
			if( (KeyIdent&0xFF) > 0x40)
				return false;

			return !Joystick[ (KeyIdent>>8)^0x80].Button[(KeyIdent&0xFF)-8];
		}
	} else
	return ((GetKeyState( KeyIdent)&0x80) == 0);
}

unsigned long S9xReadJoypad( int J)
{
char PadState[2];
JOYINFOEX jie;

	if( J > 3)
	{	return 0;	}

	if( !Joypad[ J].Enabled)
	{	return 0;	}

	// The J == 0 means that it will only read the joystick states when the first joypad
	// is read by Snes9x, the actual processing of the joystick data is done by GetState
	// But it's reported that the joyGetPosEx is pretty slow, so I only read it once...
	// This, though, works under the assumption that joypad 0 is the first joypad to be 
	// read everytime
	if( J == 0)
    for( short C = 0; C != 16; C ++)
	if( Joystick[C].Attached)
    {
        jie.dwSize = sizeof( jie);
        jie.dwFlags = JOY_RETURNBUTTONS | JOY_RETURNX | JOY_RETURNY | JOY_RETURNPOV;

        if( joyGetPosEx( JOYSTICKID1+C, &jie) != JOYERR_NOERROR)
            continue;

        if( Normalize( jie.dwXpos, Joystick[C].Caps.wXmin, Joystick[C].Caps.wXmax) < -60)
        {
            Joystick[C].Right = false;
            Joystick[C].Left = true;
        }
        else
        {   Joystick[C].Left = false;   }


        if( Normalize( jie.dwXpos, Joystick[C].Caps.wXmin, Joystick[C].Caps.wXmax) > 60)
        {
            Joystick[C].Left = false;
            Joystick[C].Right = true;
        }
        else
        {   Joystick[C].Right = false;   }

        if( Normalize( jie.dwYpos, Joystick[C].Caps.wYmin, Joystick[C].Caps.wYmax) < -60)
        {
            Joystick[C].Down = false;
            Joystick[C].Up = true;
        }
        else
        {   Joystick[C].Up = false;   }

        if( Normalize( jie.dwYpos, Joystick[C].Caps.wYmin, Joystick[C].Caps.wYmax) > 60)
        {
            Joystick[C].Up = false;
            Joystick[C].Down = true;
        }
        else
        {   Joystick[C].Down = false;   }

        switch( jie.dwPOV)
        {
            case JOY_POVBACKWARD:
                Joystick[C].PovDown = true;
                Joystick[C].PovUp = false;
                Joystick[C].PovLeft = false;
                Joystick[C].PovRight = false;
                break;

            case JOY_POVFORWARD:
                Joystick[C].PovDown = false;
                Joystick[C].PovUp = true;
                Joystick[C].PovLeft = false;
                Joystick[C].PovRight = false;
                break;

            case JOY_POVLEFT:
                Joystick[C].PovDown = false;
                Joystick[C].PovUp = false;
                Joystick[C].PovLeft = true;
                Joystick[C].PovRight = false;
                break;

            case JOY_POVRIGHT:
                Joystick[C].PovDown = false;
                Joystick[C].PovUp = false;
                Joystick[C].PovLeft = false;
                Joystick[C].PovRight = true;
                break;

            default:
                Joystick[C].PovDown = false;
                Joystick[C].PovUp = false;
                Joystick[C].PovLeft = false;
                Joystick[C].PovRight = false;
                break;
        }

        for( short B = 0; B != 32; B ++, jie.dwButtons >>= 1)
        if( (jie.dwButtons&1))
        {	Joystick[C].Button[B] = true;	}
        else
        {   Joystick[C].Button[B] = false; }
    }

	PadState[0]  = 0;
	PadState[0] |= !GetState( Joypad[ J].R)      ?  16 : 0;
	PadState[0] |= !GetState( Joypad[ J].L)      ?  32 : 0;
	PadState[0] |= !GetState( Joypad[ J].X)      ?  64 : 0;
	PadState[0] |= !GetState( Joypad[ J].A)      ? 128 : 0;

	PadState[1]  = 0;
	PadState[1] |= !GetState( Joypad[ J].Right)  ?   1 : 0;
	PadState[1] |= !GetState( Joypad[ J].Left)   ?   2 : 0;
	PadState[1] |= !GetState( Joypad[ J].Down)   ?   4 : 0;
	PadState[1] |= !GetState( Joypad[ J].Up)     ?   8 : 0;
	PadState[1] |= !GetState( Joypad[ J].Start)  ?  16 : 0;
	PadState[1] |= !GetState( Joypad[ J].Select) ?  32 : 0;
	PadState[1] |= !GetState( Joypad[ J].Y)      ?  64 : 0;
	PadState[1] |= !GetState( Joypad[ J].B)      ? 128 : 0;

	return *(WORD *)PadState | (Joypad[ J].Enabled ? 0x80000000 : 0);
}

__int16      S9xPaletteWord[256];
PALETTEENTRY S9xPaletteEntry[256];
void S9xSetPalette( void)
{
LPDIRECTDRAWPALETTE lpDDTemp;
uint16 Brightness = (IPPU.MaxBrightness) * 140;

    // Only update the palette structures if needed
	if( GUI.Depth == 8)
	{
		for( int i = 0; i < 256; i ++)
		{	
			S9xPaletteEntry[i].peRed   = (((PPU.CGDATA [i] >>  0) & 0x1F)*Brightness)>>8;
			S9xPaletteEntry[i].peGreen = (((PPU.CGDATA [i] >>  5) & 0x1F)*Brightness)>>8;
			S9xPaletteEntry[i].peBlue  = (((PPU.CGDATA [i] >> 10) & 0x1F)*Brightness)>>8;
		}

		DirectX.lpDDPalette -> SetEntries( 0, 0, 256, S9xPaletteEntry);

		if( DirectX.lpDDSPrimary != NULL)
		{
			DirectX.lpDDSPrimary -> GetPalette( &lpDDTemp);
			if( lpDDTemp != DirectX.lpDDPalette)
				DirectX.lpDDSPrimary -> SetPalette( DirectX.lpDDPalette);
		}
		else
		{
			DirectX.lpDDSPrimary2 -> GetPalette( &lpDDTemp);
			if( lpDDTemp != DirectX.lpDDPalette)
				DirectX.lpDDSPrimary2 -> SetPalette( DirectX.lpDDPalette);
		}
	}

	// This will (probably) fail on non-RGB565 cards, I'll convert it to more standards asap
	if( (GUI.Depth == 16) && (Settings.SixteenBit == false))
	{
		for( int i = 0; i < 256; i ++)
		{
			S9xPaletteWord[i]  = ((PPU.CGDATA [i] >>  0) & 0x1F) << 11;
			S9xPaletteWord[i] |= ((PPU.CGDATA [i] >>  5) & 0x1F) << 6;
			S9xPaletteWord[i] |= ((PPU.CGDATA [i] >> 10) & 0x1F) << 0;
		}
	}
}

BYTE ScreenBuffer[ 512*478*2];
BYTE SubScreenBuffer[ 512*478*2];
BYTE ScreenDepth[ 512*478];
BYTE SubScreenDepth[ 512*478];

bool8 S9xInitUpdate( void)
{
	GFX.Pitch = 512*2;
	GFX.RealPitch = 512*2;
	GFX.Screen = ScreenBuffer;
	GFX.SubScreen = SubScreenBuffer;
	GFX.DepthBuffer = ScreenDepth;
	GFX.SubDepthBuffer = SubScreenBuffer;
    GFX.SubDepthBuffer = SubScreenDepth;
	GFX.Delta = (GFX.SubScreen - GFX.Screen) >> 1;
	return 1;
}

#include "..\Render.h"

bool LockSurface( LPDIRECTDRAWSURFACE lpDDSurface, SSurface *lpSurface)
{
DDSURFACEDESC ddsd;
HRESULT hResult;
int retry;

	ddsd.dwSize = sizeof( ddsd);

	retry = 0;
	while( true)
	{
		hResult = lpDDSurface -> Lock( NULL, &ddsd, DDLOCK_WAIT, NULL);

		if( hResult == DD_OK)
		{
			lpSurface -> Width = ddsd.dwWidth;
			lpSurface -> Height = ddsd.dwHeight;
			lpSurface -> Pitch = ddsd.lPitch;
			lpSurface -> Surface = (unsigned char *)ddsd.lpSurface;
			return true;
		}

		if( hResult == DDERR_SURFACELOST)
		{
			retry++;
			if( retry > 5)
			{	return false;	}

			hResult = lpDDSurface -> Restore();
			if( hResult != DD_OK)
			{	return false;	}

			continue;
		}

		if( hResult != DDERR_WASSTILLDRAWING)
		{	return false;	}
	}
}

bool LockSurface2( LPDIRECTDRAWSURFACE2 lpDDSurface, SSurface *lpSurface)
{
DDSURFACEDESC ddsd;
HRESULT hResult;
int retry;

	ddsd.dwSize = sizeof( ddsd);

	retry = 0;
	while( true)
	{
		hResult = lpDDSurface -> Lock( NULL, &ddsd, DDLOCK_WAIT, NULL);

		if( hResult == DD_OK)
		{
			lpSurface -> Width = ddsd.dwWidth;
			lpSurface -> Height = ddsd.dwHeight;
			lpSurface -> Pitch = ddsd.lPitch;
			lpSurface -> Surface = (unsigned char *)ddsd.lpSurface;
			return true;
		}

		if( hResult == DDERR_SURFACELOST)
		{
			retry++;
			if( retry > 5)
			{	return false;	}

			hResult = lpDDSurface -> Restore();
			if( hResult != DD_OK)
			{	return false;	}

			continue;
		}

		if( hResult != DDERR_WASSTILLDRAWING)
		{	return false;	}
	}
}

bool8 S9xDeinitUpdate( int Width, int Height, bool8 sixteen_bit)
{
SSurface Src, Dst;
LPDIRECTDRAWSURFACE lpDDSurface;
LPDIRECTDRAWSURFACE2 lpDDSurface2;
bool PrimarySurfaceLockFailed = false;
RECT srcRect, dstRect;

	Src.Width = Width;
	Src.Height = Height;
	Src.Pitch = GFX.Pitch;
	Src.Surface = ScreenBuffer;

	if( DirectX.lpDDSPrimary2 == NULL)
	{
		lpDDSurface2 = NULL;
		if( DirectX.lpDDSBack == NULL)
		{	lpDDSurface = DirectX.lpDDSPrimary;	}
		else 
		{	lpDDSurface = DirectX.lpDDSBack;	}
	} else
	{
		lpDDSurface = NULL;
		if( DirectX.lpDDSBack2 == NULL)
		{	lpDDSurface2 = DirectX.lpDDSPrimary2;	}
		else 
		{	lpDDSurface2 = DirectX.lpDDSBack2;	}
	}

	if( lpDDSurface2 == NULL)
	{	
		if( !LockSurface( lpDDSurface, &Dst))
		{	// Here's the failsafe, if everything else failes, try this =)
			lpDDSurface = DirectX.lpDDSOffScreen;
			if( !LockSurface( lpDDSurface, &Dst))
			{	return false;	}

			PrimarySurfaceLockFailed = true;
		}
	}
	else
	{	
		if( !LockSurface2( lpDDSurface2, &Dst))
		{	// Here's the failsafe, if everything else failes, try this =)
			lpDDSurface2 = DirectX.lpDDSOffScreen2;
			if( !LockSurface2( lpDDSurface2, &Dst))
			{	return false;	}
			
			PrimarySurfaceLockFailed = true;
		}
	}

	// This call should be moved to a more approriate place
	SelectRenderMethod();

	RenderMethod( Src, Dst);

	if( PrimarySurfaceLockFailed)
	{
		if( (GUI.Scale == 0) || (Width == 512) || (Height > 239))
		{
			srcRect.top    = (Dst.Height  - ((Height > 239) ? 478 : 239)) >> 1;
			srcRect.left   = (Dst.Width   - Width) >> 1;
			srcRect.bottom = srcRect.top  + ((Height > 239) ? 478 : 239);
			srcRect.right  = srcRect.left + Width;
		}
		else
		{
			srcRect.top    = (Dst.Height  - 512) >> 1;
			srcRect.left   = (Dst.Width   - 478) >> 1;
			srcRect.bottom = srcRect.top  + 478;
			srcRect.right  = srcRect.left + 512;
		}

		dstRect.top    = (DirectX.Height - (srcRect.bottom-srcRect.top)) >> 1;
		dstRect.left   = (DirectX.Width  - (srcRect.right -srcRect.left)) >> 1;
		dstRect.bottom = dstRect.top     + (srcRect.bottom-srcRect.top);
		dstRect.right  = dstRect.left    + (srcRect.right -srcRect.left);
	}
	
	if( lpDDSurface2 == NULL)
	{	
		lpDDSurface -> Unlock( NULL);
		if( PrimarySurfaceLockFailed)
		{
			if( DirectX.lpDDSBack == NULL)
			{	lpDDSurface = DirectX.lpDDSPrimary;	}
			else 
			{	lpDDSurface = DirectX.lpDDSBack;	}

			if( lpDDSurface -> Blt( &dstRect, DirectX.lpDDSOffScreen, &srcRect, DDBLT_WAIT, NULL) != DD_OK)
			{	return false;	}
		}
		DirectX.lpDDSPrimary -> Flip( NULL, DDFLIP_WAIT);
	}
	else
	{	
		lpDDSurface2 -> Unlock( NULL);	
		if( PrimarySurfaceLockFailed)
		{
			if( DirectX.lpDDSBack2 == NULL)
			{	lpDDSurface2 = DirectX.lpDDSPrimary2;	}
			else 
			{	lpDDSurface2 = DirectX.lpDDSBack2;	}

			if( lpDDSurface2 -> Blt( &dstRect, DirectX.lpDDSOffScreen2, &srcRect, DDBLT_WAIT, NULL) != DD_OK)
			{	return false;	}
		}
		DirectX.lpDDSPrimary2 -> Flip( NULL, DDFLIP_WAIT);
	}
	return true;
}

void InitSnes9X( void)
{
#ifdef DEBUGGER
    extern FILE *trace;

    trace = fopen( "SNES9X.TRC", "wt");
    freopen( "SNES9X.OUT", "wt", stdout);

    //CPU.Flags |= TRACE_FLAG;
    //APU.Flags |= TRACE_FLAG;
#endif

    // CPU options 
    Settings.APUEnabled = true;
    Settings.Shutdown = true;
    Settings.SoundSkipMethod = 0;
    Settings.H_Max = (100 * SNES_CYCLES_PER_SCANLINE) / 100;
    Settings.HBlankStart = (256 * Settings.H_Max) / SNES_HCOUNTER_MAX;

    Settings.SPCTo65c816Ratio = 2;
    Settings.DisableIRQ = false;
    Settings.Paused = false;

    // Tracing options
    Settings.TraceDMA = false;
    Settings.TraceHDMA = false;
    Settings.TraceVRAM = false;
    Settings.TraceUnknownRegisters = false;
    Settings.TraceDSP = false;

    // Joystick options
    Settings.SwapJoypads = false;
    Settings.JoystickEnabled = false;

    // ROM timing options (see also H_Max above)
    Settings.PAL = false;
    Settings.FrameTimePAL = 20;
    Settings.FrameTimeNTSC = 17;
    Settings.FrameTime = 17;

    // ROM image and peripheral options
    Settings.ForceSuperFX = false;
    Settings.ForceNoSuperFX = false;
    Settings.MultiPlayer5 = false;
    Settings.Mouse = false;
    Settings.SuperScope = false;
    Settings.MultiPlayer5Master = false;
    Settings.SuperScopeMaster = false;
    Settings.MouseMaster = false;
    Settings.SuperFX = true;

    // Sound options
    Settings.TraceSoundDSP = false;
    Settings.SoundPlaybackRate = 7;
    Settings.Stereo = _stereo;
    Settings.SoundBufferSize = _bytecount;
    Settings.SoundEnvelopeHeightReading = false;
    Settings.DisableSoundEcho = false;
    Settings.SoundSync = true;
    Settings.InterpolatedSound = true;

    // SNES graphics options
    Settings.BGLayering = false;
    Settings.DisableGraphicWindows = false;
    Settings.ForceTransparency = false;
    Settings.ForceNoTransparency = false;
    Settings.DisableHDMA = false;

    Memory.Init();

    GFX.Pitch = 512*2;
    GFX.RealPitch = 512*2;
    GFX.Screen = ScreenBuffer;
    GFX.SubScreen = SubScreenBuffer;
    GFX.Delta = (GFX.SubScreen - GFX.Screen) >> 1;

    if((DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
       DirectX.DDPixelFormat.dwRGBBitCount == 16 &&
       DirectX.DDPixelFormat.dwRBitMask == 0xF800 &&
       DirectX.DDPixelFormat.dwGBitMask == 0x07E0 &&
       DirectX.DDPixelFormat.dwBBitMask == 0x001F)
        S9xSetRenderPixelFormat( RGB565);

    if( (DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
        (DirectX.DDPixelFormat.dwRGBBitCount == 16 ||
         DirectX.DDPixelFormat.dwRGBBitCount == 15)&&
        DirectX.DDPixelFormat.dwRBitMask == 0x7C00 &&
        DirectX.DDPixelFormat.dwGBitMask == 0x03E0 &&
        DirectX.DDPixelFormat.dwBBitMask == 0x001F)
        S9xSetRenderPixelFormat( RGB555);

    if((DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
       DirectX.DDPixelFormat.dwRGBBitCount == 16 &&
       DirectX.DDPixelFormat.dwRBitMask == 0x001F &&
       DirectX.DDPixelFormat.dwGBitMask == 0x07E0 &&
       DirectX.DDPixelFormat.dwBBitMask == 0xF800)
        S9xSetRenderPixelFormat( BGR565);

    if( (DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
        (DirectX.DDPixelFormat.dwRGBBitCount == 16 ||
         DirectX.DDPixelFormat.dwRGBBitCount == 15)&&
        DirectX.DDPixelFormat.dwRBitMask == 0x001F &&
        DirectX.DDPixelFormat.dwGBitMask == 0x03E0 &&
        DirectX.DDPixelFormat.dwBBitMask == 0x7C00)
        S9xSetRenderPixelFormat( BGR555);

    S9xGraphicsInit();

    S9xInitAPU();
    S9xInitSound( 7, true, _bytecount);

    so.playback_rate = _rate;
    so.stereo = _stereo;
    so.sixteen_bit = _sixteen_bit;
    so.buffer_size = _bytecount;

    for( short C = 0; C != 16; C ++)
        Joystick[C].Attached = joyGetDevCaps( JOYSTICKID1+C, &Joystick[C].Caps, sizeof( JOYCAPS)) == JOYERR_NOERROR;

    GUI.StopEmulation = !Memory.LoadROM( "SNES9X.SMC");
    if( !GUI.StopEmulation)	Memory.LoadSRAM( S9xGetSRAMFilename());
}
