/*****************************************************************************/
/*  Snes9X: Win32                                                            */
/*****************************************************************************/

#include "Snes9X.h"
#include "DirectX.h"
#include "Snes9X/Snes9X.H"
#include "Snes9X/MemMap.H"

/*****************************************************************************/
/* Global variables                                                          */
/*****************************************************************************/
const char Key[] = "Software\\Emulators\\Snes9X\\1.19a";
struct sGUI GUI;

CDirectX DirectX;
struct SJoypad Joypad[4] = {
	{ true,								/* Joypad 1 default keys:*/
	VK_LEFT, VK_RIGHT, VK_UP, VK_DOWN,	/* Left, Right, Up, Down */
	VK_RETURN, VK_SPACE,					/* Start, Select */
	'D', 'C',							/* A B */
	'S', 'X',							/* X Y */
	'A', 'Z'							/* L R */
	}, { true,							/* Joypad 2 default joystick 0 */
	0x8000, 0x8001, 0x8002, 0x8003,		/* Left, Right, Up, Down */
	0x8008, 0x8009,						/* Start, Select */
	0x800A, 0x800B,						/* A B */
	0x800C, 0x800D,						/* X Y */
	0x800E, 0x800F						/* L R */
	}, { false,							/* Joypad 3 default off */
	0, 0, 0, 0,							/* Left, Right, Up, Down */
	0, 0,								/* Start, Select */
	0, 0,								/* A B */
	0, 0,								/* X Y */
	0, 0								/* L R */
	}, { false,							/* Joypad 4 default off */
	0, 0, 0, 0,							/* Left, Right, Up, Down */
	0, 0,								/* Start, Select */
	0, 0,								/* A B */
	0, 0,								/* X Y */
	0, 0								/* L R */
	}};

struct SSoundRates 
{
	long rate;
	bool stereo;
	bool sixteen_bit;
	int ident;
} SoundRates[16] = {
	{  8000, false, false, ID_SOUND_8BIT8000HZMONO},
	{ 11000, false, false, ID_SOUND_8BIT11025HZMONO},
	{ 22000, false, false, ID_SOUND_8BIT22050HZMONO},
	{ 44000, false, false, ID_SOUND_8BIT44100HZMONO},
	{  8000, false, true , ID_SOUND_16BIT8000HZMONO},
	{ 11000, false, true , ID_SOUND_16BIT11025HZMONO},
	{ 22000, false, true , ID_SOUND_16BIT22050HZMONO},
	{ 44000, false, true , ID_SOUND_16BIT44100HZMONO},
	{  8000,  true, false, ID_SOUND_8BIT8000HZSTEREO},
	{ 11000,  true, false, ID_SOUND_8BIT11025HZSTEREO},
	{ 22000,  true, false, ID_SOUND_8BIT22050HZSTEREO},
	{ 44000,  true, false, ID_SOUND_8BIT44100HZSTEREO},
	{  8000,  true, true , ID_SOUND_16BIT8000HZSTEREO},
	{ 11000,  true, true , ID_SOUND_16BIT11025HZSTEREO},
	{ 22000,  true, true , ID_SOUND_16BIT22050HZSTEREO},
	{ 44000,  true, true , ID_SOUND_16BIT44100HZSTEREO}};

// Languages supported by Snes9X: Windows
// 0 - English [Default]
// 1 - Dutch/Nederlands
struct sLanguages Languages[] = {
	{ IDR_MENU_US, 
	  "DirectX failed to initialize!", 
	  "DirectDraw failed to set the selected display mode!", 
	  "DirectSound failed to initialize, no sound will be played.",
	  "These settings won't take effect until you restart the emulator.",
	  "The frame timer failed to initialize, please do NOT select the automatic framerate option or Snes9X will crash!"},
	{ IDR_MENU_NL, 
	  "Er is een fout opgetreden tijdens het initalizeren van DirectX!", 
	  "Er is een fout opgetreden tijdens het verander van scherm modus!", 
	  "Er is een fout opgetreden tijdens het initializeren van DirectSound, er zal geen geluid worden afgespeeld.",
	  "Deze opties worden pas toegepast als de emulator opnieuw is opgestart.",
	  "Er is een fout opgetreden tijdens het initializeren van de frame timer, kies NIET de automatische framerate optie want dan zal Snes9X crashen!"}};

extern "C" {
	bool __cdecl _OpenROM( char *, struct SSettings *);
	void __cdecl _JoypadConfig( struct SJoypad *J);
	void __cdecl _InitDLL( HANDLE hWnd);
	void __cdecl _DirectXConfig( LPDIRECTDRAW plpDD, struct SSettings *Settings, struct sGUI *GUI, bool *showFPS);
};

/*****************************************************************************/
/* WinProc                                                                   */
/*****************************************************************************/
bool SetupSound( long rate, bool sixteen_bit, bool stereo);
#include "snes9x/ppu.h"
#include "snes9x/snapshot.h"
extern "C" const char *S9xGetSRAMFilename ();

void SwitchToGDI()
{
	IPPU.ColorsChanged = true;
	if( DirectX.lpDDSPrimary != NULL)
	{
		DirectX.lpDD -> FlipToGDISurface();
		DirectX.lpDDSPrimary -> SetPalette( NULL);
	}
	else
	{
		DirectX.lpDD -> FlipToGDISurface();
		DirectX.lpDDSPrimary2 -> SetPalette( NULL);
	}
}

void UpdateBackBuffer()
{
DDSURFACEDESC ddsd;
HRESULT hResult;
int retry;
RECT Src;

	SwitchToGDI();

	ddsd.dwSize = sizeof( ddsd);

	Src.left = 0; Src.right = DirectX.Width;
	Src.top = 0; Src.bottom = DirectX.Height;

	retry = 0;
	while( true)
	{
		if( DirectX.lpDDSPrimary2 != NULL)
		{	hResult = DirectX.lpDDSPrimary2 -> Lock( NULL, &ddsd, DDLOCK_WAIT, NULL);	}
		else
		{	hResult = DirectX.lpDDSPrimary -> Lock( NULL, &ddsd, DDLOCK_WAIT, NULL);	}

		if( hResult == DD_OK)
		{
			unsigned char *B = (unsigned char *)ddsd.lpSurface;
			for( int H = 0; H != DirectX.Height; H ++, B += ddsd.lPitch)
			if( DirectX.Depth == 8)
			{	ZeroMemory( B, DirectX.Width);	}
			else
			{	ZeroMemory( B, DirectX.Width*2);	}


			if( DirectX.lpDDSPrimary2 != NULL)
			{	DirectX.lpDDSPrimary2 -> Unlock( NULL);	}
			else
			{	DirectX.lpDDSPrimary -> Unlock( NULL);	}
		}

		if( hResult == DDERR_SURFACELOST)
		{
			retry++;
			if( retry > 5)
			{	break;	}

			if( DirectX.lpDDSPrimary2 != NULL)
			{	hResult = DirectX.lpDDSPrimary2 -> Restore();	}
			else
			{	hResult = DirectX.lpDDSPrimary -> Restore();	}
			
			if( hResult != DD_OK)
			{	break;	}

			continue;
		}

		if( hResult != DDERR_WASSTILLDRAWING)
		{	break;	}
	}

	if( GetMenu( GUI.hWnd) != NULL)
	{	DrawMenuBar( GUI.hWnd);	}

	if( DirectX.lpDDSPrimary2 != NULL)
	{	DirectX.lpDDSBack2 -> BltFast( 0, 0, DirectX.lpDDSPrimary2, &Src, DDBLTFAST_WAIT);	}
	else
	{	DirectX.lpDDSBack -> BltFast( 0, 0, DirectX.lpDDSPrimary, &Src, DDBLTFAST_WAIT);	}
}

LRESULT CALLBACK WinProc(
	HWND hWnd,
	UINT uMsg,
	WPARAM wParam,
	LPARAM lParam)
{
int i;
MENUITEMINFO mii;
bool showFPS;
	switch( uMsg)
    {
	case WM_KEYDOWN:
		switch( wParam)
		{
		case VK_ESCAPE:
			if( GetMenu( GUI.hWnd) == NULL)
			{	SetMenu( GUI.hWnd, GUI.hMenu);	}
			else
			{	SetMenu( GUI.hWnd, NULL);	}

			UpdateBackBuffer();
			break;

		case '0':
			Settings.DisableHDMA = !Settings.DisableHDMA;
			break;
		case '1':
			PPU.BG_Forced ^= 1;
			break;
		case '2':
			PPU.BG_Forced ^= 2;
			break;
		case '3':
			PPU.BG_Forced ^= 4;
			break;
		case '4':
			PPU.BG_Forced ^= 8;
			break;
		case '5':
			PPU.BG_Forced ^= 16;
			break;
		case '6':
			Settings.SwapJoypads = !Settings.SwapJoypads;
			break;
		case '8':
			Settings.BGLayering = !Settings.BGLayering;
			break;
		case '9':
			if( Settings.SixteenBit)
			{	Settings.Transparency = !Settings.Transparency;	}
			break;

		case VK_BACK:
			Settings.DisableGraphicWindows = !Settings.DisableGraphicWindows;
			break;

		case VK_PAUSE:
			GUI.PauseEmulation = GUI.PauseEmulation ^ true;
			break;

		case VK_F1:
		case VK_F2:
		case VK_F3:
		case VK_F4:
		case VK_F5:
		case VK_F6:
		case VK_F7:
		case VK_F8:
		case VK_F9:
			{
		    char filename [_MAX_PATH];
			char drive [_MAX_DRIVE];
			char dir [_MAX_DIR];
			char fname [_MAX_FNAME];
			char ext [_MAX_EXT];

		    _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
			strcpy( ext, ".000");
			ext[4] += wParam-VK_F1;
			_makepath( filename, drive, dir, fname, ext);

			GUI.ForcedPauseEmulation = true;

			if( (GetKeyState( VK_SHIFT)&0x80) == 0)
			{	S9xUnfreezeGame( filename);	}
			else
			{	S9xFreezeGame( filename);	}

			UpdateBackBuffer();
			GUI.ForcedPauseEmulation = false;

			}
			break;
		}
		break;

	case WM_COMMAND:
		switch( wParam)
		{
		case ID_OPTIONS_DISPLAY:
			GUI.ForcedPauseEmulation = true;

			if( !DirectX.SetDisplayMode( 640, 480, 0, false, false))
			{
				MessageBox( GUI.hWnd, Languages[ GUI.Language].errModeDD, "Snes9X - DrawDraw", MB_OK | MB_ICONSTOP);
				return false; 
			}

			SwitchToGDI();
			showFPS = Settings.DisplayFrameRate ? true : false;
			_DirectXConfig( DirectX.lpDD, &Settings, &GUI, &showFPS);
			Settings.DisplayFrameRate = showFPS;

			if( !DirectX.SetDisplayMode( GUI.Width, GUI.Height, GUI.Depth, false, GUI.DoubleBuffered))
			{
				MessageBox( GUI.hWnd, Languages[ GUI.Language].errModeDD, "Snes9X - DrawDraw", MB_OK | MB_ICONSTOP);
				return false; 
			}

			S9xGraphicsDeinit();

			if((DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
				DirectX.DDPixelFormat.dwRGBBitCount == 16 &&
				DirectX.DDPixelFormat.dwRBitMask == 0xF800 &&
				DirectX.DDPixelFormat.dwGBitMask == 0x07E0 &&
				DirectX.DDPixelFormat.dwBBitMask == 0x001F)
				S9xSetRenderPixelFormat( RGB565);

			if( (DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
				(DirectX.DDPixelFormat.dwRGBBitCount == 16 ||
				DirectX.DDPixelFormat.dwRGBBitCount == 15)&&
				DirectX.DDPixelFormat.dwRBitMask == 0x7C00 &&
				DirectX.DDPixelFormat.dwGBitMask == 0x03E0 &&
				DirectX.DDPixelFormat.dwBBitMask == 0x001F)
				S9xSetRenderPixelFormat( RGB555);

			if((DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
				DirectX.DDPixelFormat.dwRGBBitCount == 16 &&
				DirectX.DDPixelFormat.dwRBitMask == 0x001F &&
				DirectX.DDPixelFormat.dwGBitMask == 0x07E0 &&
				DirectX.DDPixelFormat.dwBBitMask == 0xF800)
				S9xSetRenderPixelFormat( BGR565);

			if( (DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
				(DirectX.DDPixelFormat.dwRGBBitCount == 16 ||
				DirectX.DDPixelFormat.dwRGBBitCount == 15)&&
				DirectX.DDPixelFormat.dwRBitMask == 0x001F &&
				DirectX.DDPixelFormat.dwGBitMask == 0x03E0 &&
				DirectX.DDPixelFormat.dwBBitMask == 0x7C00)
				S9xSetRenderPixelFormat( BGR555);

			S9xGraphicsInit();

			IPPU.RenderThisFrame = false;

			UpdateBackBuffer();
			GUI.ForcedPauseEmulation = false;
			break;
		case ID_OPTIONS_JOYPAD:
			GUI.ForcedPauseEmulation = true;

			if( !DirectX.SetDisplayMode( 640, 480, 0, false, false))
			{
				MessageBox( GUI.hWnd, Languages[ GUI.Language].errModeDD, "Snes9X - DrawDraw", MB_OK | MB_ICONSTOP);
				return false; 
			}

			SwitchToGDI();
			_JoypadConfig( Joypad);

			if( !DirectX.SetDisplayMode( GUI.Width, GUI.Height, GUI.Depth, false, GUI.DoubleBuffered))
			{
				MessageBox( GUI.hWnd, Languages[ GUI.Language].errModeDD, "Snes9X - DrawDraw", MB_OK | MB_ICONSTOP);
				return false; 
			}

			UpdateBackBuffer();
			GUI.ForcedPauseEmulation = false;
			break;

		case ID_FILE_OPEN:
			{
			char FileName[ _MAX_PATH];
			char S[80];
				
				GUI.ForcedPauseEmulation = true;

				if( !DirectX.SetDisplayMode( 640, 480, 0, false, false))
				{
					MessageBox( GUI.hWnd, Languages[ GUI.Language].errModeDD, "Snes9X - DrawDraw", MB_OK | MB_ICONSTOP);
					return false; 
				}

				SwitchToGDI();

			    extern FILE *trace_fs;
				if( _OpenROM( FileName, &Settings))
				{
			        fprintf (trace_fs, "Loading \"%s\"\n", FileName);
			        fflush (trace_fs);

					if( !GUI.StopEmulation) Memory.SaveSRAM( S9xGetSRAMFilename());
					GUI.StopEmulation = !Memory.LoadROM( FileName);
					if( !GUI.StopEmulation) Memory.LoadSRAM( S9xGetSRAMFilename());

					sprintf( S, "Loaded \"%s\"", FileName);
					DirectX.OnScreenMsg( S, 2000);
				}

				UpdateBackBuffer();
				GUI.ForcedPauseEmulation = false;

				if( !DirectX.SetDisplayMode( GUI.Width, GUI.Height, GUI.Depth, false, GUI.DoubleBuffered))
				{
					MessageBox( GUI.hWnd, Languages[ GUI.Language].errModeDD, "Snes9X - DrawDraw", MB_OK | MB_ICONSTOP);
					return false; 
				}
			}
			break;

		case ID_FILE_EXIT:
			PostMessage( hWnd, WM_DESTROY, 0, 0);
			break;

		case ID_WINDOW_HIDEMENUBAR:
			if( GetMenu( GUI.hWnd) == NULL)
			{	SetMenu( GUI.hWnd, GUI.hMenu);	}
			else
			{	SetMenu( GUI.hWnd, NULL);	}
			break;

		case ID_LANGUAGE_ENGLISH:
			GUI.Language = 0;

			SetMenu( GUI.hWnd, LoadMenu( GUI.hInstance, MAKEINTRESOURCE( Languages[ GUI.Language].idMenu)));
			DestroyMenu( GUI.hMenu);
			GUI.hMenu = GetMenu( GUI.hWnd);
			break;
		case ID_LANGUAGE_NEDERLANDS:
			GUI.Language = 1;

			SetMenu( GUI.hWnd, LoadMenu( GUI.hInstance, MAKEINTRESOURCE( Languages[ GUI.Language].idMenu)));
			DestroyMenu( GUI.hMenu);
			GUI.hMenu = GetMenu( GUI.hWnd);
			break;

		case ID_SOUND_8BIT8000HZMONO:
		case ID_SOUND_8BIT11025HZMONO:
		case ID_SOUND_8BIT22050HZMONO:
		case ID_SOUND_8BIT44100HZMONO:
		case ID_SOUND_16BIT8000HZMONO:
		case ID_SOUND_16BIT11025HZMONO:
		case ID_SOUND_16BIT22050HZMONO:
		case ID_SOUND_16BIT44100HZMONO:
		case ID_SOUND_8BIT8000HZSTEREO:
		case ID_SOUND_8BIT11025HZSTEREO:
		case ID_SOUND_8BIT22050HZSTEREO:
		case ID_SOUND_8BIT44100HZSTEREO:
		case ID_SOUND_16BIT8000HZSTEREO:
		case ID_SOUND_16BIT11025HZSTEREO:
		case ID_SOUND_16BIT22050HZSTEREO:
		case ID_SOUND_16BIT44100HZSTEREO:
			GUI.ForcedPauseEmulation = true;

			for( i = 0; i != 16; i ++)
			if( SoundRates[i].ident == (int) wParam)
			{
				GUI.SoundMode = i;
				if( !SetupSound( SoundRates[GUI.SoundMode].rate, SoundRates[GUI.SoundMode].sixteen_bit, SoundRates[GUI.SoundMode].stereo))
				{	MessageBox( GUI.hWnd, Languages[ GUI.Language].errInitDS, "Snes9X - DrawSound", MB_OK | MB_ICONINFORMATION);	}
				break;
			}

			UpdateBackBuffer();
			GUI.ForcedPauseEmulation = false;
			break;

		case ID_SOUND_NOSOUND:
			GUI.ForcedPauseEmulation = true;

			GUI.SoundMode = -1;
			SetupSound( 0, false, false);

			UpdateBackBuffer();
			GUI.ForcedPauseEmulation = false;
			break;
		}
		break;


	case WM_EXITMENULOOP:
		UpdateBackBuffer();
		GUI.ForcedPauseEmulation = false;
		break;

	case WM_ENTERMENULOOP:
		ZeroMemory( &mii, sizeof( mii));
		mii.cbSize = sizeof( mii);
		mii.fMask = MIIM_STATE;

		mii.fState = MFS_UNCHECKED;
		if( !DirectX.DSAvailable)
		{	mii.fState = MFS_DISABLED;	}

		SetMenuItemInfo( GUI.hMenu, ID_SOUND_NOSOUND, FALSE, &mii);
		for( i = 0; i != 16; i++)
		{	SetMenuItemInfo( GUI.hMenu, SoundRates[i].ident, FALSE, &mii);	}
		
		mii.fState = !DirectX.DSAvailable ? MFS_CHECKED | MFS_DISABLED : MFS_CHECKED;

		if( !DirectX.DSAvailable || GUI.SoundMode == -1)
		{	SetMenuItemInfo( GUI.hMenu, ID_SOUND_NOSOUND, FALSE, &mii);	}
		else
		{	SetMenuItemInfo( GUI.hMenu, SoundRates[GUI.SoundMode].ident, FALSE, &mii);	}

		GUI.ForcedPauseEmulation = true;
		SwitchToGDI();
		DrawMenuBar( GUI.hWnd);
		break;

	case WM_DESTROY:
		PostQuitMessage( 0);
		break;
	}
	return DefWindowProc( hWnd, uMsg, wParam, lParam);
}

/*****************************************************************************/
/* WinInit                                                                   */
/*****************************************************************************/
BOOL WinInit( HINSTANCE hInstance)
{
WNDCLASS wndclass;

	wndclass.style = CS_HREDRAW|CS_VREDRAW;
	wndclass.lpfnWndProc = WinProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon( NULL, IDI_WINLOGO);
	wndclass.hCursor = LoadCursor( NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)GetStockObject( BLACK_BRUSH);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = "Snes9X: WndClass";

	GUI.hInstance = hInstance;

	if( !RegisterClass( &wndclass))
		return FALSE;
	
	if( (GUI.hMenu = LoadMenu( hInstance, MAKEINTRESOURCE( Languages[ GUI.Language].idMenu))) == NULL)
		return FALSE;

	if( (GUI.hWnd = CreateWindow(
		"Snes9X: WndClass",
		"Snes9X: Windows",
		WS_POPUPWINDOW|WS_CAPTION|WS_THICKFRAME|WS_VISIBLE,
		0, 0,
		264+256, 286+239,
		NULL,
		GUI.hMenu,
		hInstance,
		NULL)) == NULL)
		return FALSE;

	return TRUE;
}

/*****************************************************************************/
/* WinSave/WinLoad - Save/Load the settings to/from the registry             */
/*****************************************************************************/
#define SetKey( n, v)\
	RegSetValueEx( hKey, n, 0, REG_BINARY, (const unsigned char *) &v, sizeof( v));
#define GetKey( n, v)\
	cbData = sizeof( v);\
	RegQueryValueEx( hKey, n, 0, NULL, (unsigned char *) &v, &cbData);
#define GetKeyDef( n, v, d)\
	cbData = sizeof( v);\
	if( RegQueryValueEx( hKey, n, 0, NULL, (unsigned char *) &v, &cbData) != ERROR_SUCCESS)\
		v = d;

void WinSave( void)
{
HKEY hKey;

	if( RegCreateKeyEx( HKEY_CURRENT_USER, Key, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, NULL) == ERROR_SUCCESS)
	{
		SetKey( "Language", GUI.Language);
		SetKey( "Sound Mode", GUI.SoundMode);
		SetKey( "Stereo Sound", GUI.Stereo);


		SetKey( "Width", GUI.Width);
		SetKey( "Height", GUI.Height);
		SetKey( "Depth", GUI.Depth);
		SetKey( "Scale", GUI.Scale);
		SetKey( "DoubleBuffered", GUI.DoubleBuffered);
		SetKey( "Show FPS", Settings.DisplayFrameRate);

		SetKey( "Force LoROM", Settings.ForceLoROM);
		SetKey( "Force HiROM", Settings.ForceHiROM);
		SetKey( "Force Not Interleaved", Settings.ForceNotInterleaved);
		SetKey( "Force Interleaved Mode #1", Settings.ForceInterleaved);
		SetKey( "Force Interleaved Mode #2", Settings.ForceInterleaved2);
		SetKey( "Force PAL", Settings.ForcePAL);
		SetKey( "Force NTSC", Settings.ForceNTSC);
		SetKey( "Force Header", Settings.ForceHeader);
		SetKey( "Force NoHeader", Settings.ForceNoHeader);

		SetKey( "Render SixteenBit", Settings.SixteenBit);
		SetKey( "Enable Transparency", Settings.Transparency);
		SetKey( "Support HiRes", Settings.SupportHiRes);
		SetKey( "Skip Frames", Settings.SkipFrames);

		SetKey( "Joypad #1", Joypad[0]);
		SetKey( "Joypad #2", Joypad[1]);
		SetKey( "Joypad #3", Joypad[2]);
		SetKey( "Joypad #4", Joypad[3]);
		RegCloseKey( hKey);
	}
}

void WinLoad( void)
{
HKEY hKey;
unsigned long cbData;

	if( RegOpenKeyEx( HKEY_CURRENT_USER, Key, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
	{
		GetKey( "Language", GUI.Language);
		GetKey( "Sound Mode", GUI.SoundMode);
		GetKey( "Stereo Sound", GUI.Stereo);

		GetKeyDef( "Width", GUI.Width, 640);
		GetKeyDef( "Height", GUI.Height, 480);
		GetKeyDef( "Depth", GUI.Depth, 16);
		GetKeyDef( "Scale", GUI.Scale, 0);
		GetKeyDef( "DoubleBuffered", GUI.DoubleBuffered, false);
		GetKeyDef( "Show FPS", Settings.DisplayFrameRate, false);

		GetKeyDef( "Force LoROM", Settings.ForceLoROM, false);
		GetKeyDef( "Force HiROM", Settings.ForceHiROM, false);

		GetKeyDef( "Force Not Interleaved", Settings.ForceNotInterleaved, false);
		GetKeyDef( "Force Interleaved Mode #1", Settings.ForceInterleaved, false);
		GetKeyDef( "Force Interleaved Mode #2", Settings.ForceInterleaved2, false);

		GetKeyDef( "Force PAL", Settings.ForcePAL, false);
	    GetKeyDef( "Force NTSC", Settings.ForceNTSC, false);
	    GetKeyDef( "Force Header", Settings.ForceHeader, false);
	    GetKeyDef( "Force NoHeader", Settings.ForceNoHeader, false);

	    GetKeyDef( "Render SixteenBit", Settings.SixteenBit, true);
	    GetKeyDef( "Enable Transparency", Settings.Transparency, true);
	    GetKeyDef( "Support HiRes", Settings.SupportHiRes, true);
	    GetKeyDef( "Skip Frames", Settings.SkipFrames, AUTO_FRAMERATE);

		GetKey( "Joypad #1", Joypad[0]);
		GetKey( "Joypad #2", Joypad[1]);
		GetKey( "Joypad #3", Joypad[2]);
		GetKey( "Joypad #4", Joypad[3]);
		RegCloseKey( hKey);
	} 
	else 
	{
		GUI.Width = 640;
		GUI.Height = 480;
		GUI.Depth = 16;
		GUI.Scale = 0;
		GUI.DoubleBuffered = false;
		Settings.DisplayFrameRate = false;

		Settings.ForceLoROM = false;
		Settings.ForceInterleaved = false;

		Settings.ForceNotInterleaved = false;
		Settings.ForceInterleaved = false;
		Settings.ForceInterleaved2 = false;

		Settings.ForcePAL = false;
		Settings.ForceNTSC = false;
		Settings.ForceHeader = false;
		Settings.ForceNoHeader = false;

		Settings.SixteenBit = true;
		Settings.Transparency = true;
		Settings.SupportHiRes = true;
		Settings.SkipFrames = AUTO_FRAMERATE;
	}
}

extern long _rate;
extern bool _sixteen_bit;
extern bool _stereo;
extern long _buffernos;
extern long _buffersize;
extern long _samplecount;
extern long _bytecount;
extern unsigned long _interval;

volatile long CurrentFrame = 0;
volatile long TargetFrame = 0;
volatile unsigned long LastSound = 0;
VOID CALLBACK FrameTimer( UINT idEvent, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
	if( GUI.hFrameTimer != idEvent)
	{	return;	}

	while( (timeGetTime() - LastSound) >= _interval)
	{
		LastSound += _interval;

		void ProcessSound( void);
		ProcessSound();
	}

	if( timeGetTime() < GUI.FrameTimerStart)
	{	return;	}

	GUI.FrameTimerStart = GUI.FrameTimerStart + Settings.FrameTime;

	if( GUI.PauseEmulation || GUI.StopEmulation || GUI.ForcedPauseEmulation)
	{	return;	}

	TargetFrame++;
	SetEvent( GUI.hDrawFrame);	
	return;
}

/*****************************************************************************/
/* WinMain                                                                   */
/*****************************************************************************/
extern "C" void S9xMainLoop(void);
int WINAPI WinMain( 
	HINSTANCE hInstance,
	HINSTANCE hPrevInstance,
	LPSTR lpCmdLine,
	int nCmdShow)
{
MSG msg;
DWORD wSoundTimerRes;

	WinLoad();
	WinInit( hInstance);
	_InitDLL( GUI.hWnd);

	if( !DirectX.Initialize())
	{
		MessageBox( GUI.hWnd, Languages[ GUI.Language].errInitDD, "Snes9X - DirectX", MB_OK | MB_ICONSTOP);
		return false; 
	}
	
	if( !DirectX.SetDisplayMode( GUI.Width, GUI.Height, GUI.Depth, false, GUI.DoubleBuffered))
	{
		MessageBox( GUI.hWnd, Languages[ GUI.Language].errModeDD, "Snes9X - DrawDraw", MB_OK | MB_ICONSTOP);
		return false; 
	}

	if( GUI.SoundMode == -1)
	{	SetupSound( 0, false, false);	}
	else
	if( !SetupSound( SoundRates[GUI.SoundMode].rate, SoundRates[GUI.SoundMode].sixteen_bit, SoundRates[GUI.SoundMode].stereo))
	{	MessageBox( GUI.hWnd, Languages[ GUI.Language].errInitDS, "Snes9X - DrawSound", MB_OK | MB_ICONINFORMATION);	}

	GUI.FrameTimerStart = timeGetTime();
	
	TIMECAPS tc;
	if( timeGetDevCaps(&tc, sizeof(TIMECAPS))== TIMERR_NOERROR)
	{
		wSoundTimerRes = min(max(tc.wPeriodMin, 1), tc.wPeriodMax);
		timeBeginPeriod( wSoundTimerRes);
	}

	GUI.StopEmulation = true;
	GUI.hFrameTimer = timeSetEvent( 1, 0, &FrameTimer, 0, TIME_PERIODIC);
	if( GUI.hFrameTimer)
	{
		GUI.hDrawFrame = CreateEvent( NULL, FALSE, FALSE, NULL);
		if( GUI.hDrawFrame == NULL)
		{
			timeKillEvent( GUI.hFrameTimer);
			GUI.hFrameTimer = 0;
		}
	}
	if( GUI.hFrameTimer == NULL)
	{
		MessageBox( GUI.hWnd, Languages[ GUI.Language].errFrameTimer, "Snes9X - Frame Timer", MB_OK | MB_ICONINFORMATION);
	}

	void InitSnes9X( void);
	InitSnes9X();

	while( TRUE)
	{
		if( PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE))
		{
			if( !GetMessage( &msg, NULL, 0, 0))
				break;

			TranslateMessage( &msg);
			DispatchMessage( &msg);
		}

		if( !GUI.StopEmulation && !GUI.PauseEmulation && !GUI.ForcedPauseEmulation)
			S9xMainLoop();
	}

	if( GUI.hFrameTimer)
	{	
		timeKillEvent( GUI.hFrameTimer);
		timeEndPeriod( wSoundTimerRes);
		CloseHandle( GUI.hDrawFrame);
	}

	if( !GUI.StopEmulation) Memory.SaveSRAM( S9xGetSRAMFilename());
	WinSave();
	return msg.wParam;
}

/*****************************************************************************/
